#!/usr/bin/env Python 
"""The implementation of an HLA molecules  
"""
# load the modules
from __future__ import annotations 
import mhcnames
# define the class 
class HLAChain: 
    def __init__(self, name:str)->HLAChain:
        """Create an HLAChain instance. 
        :param name: the allele name 
        :type name: str
        :return: an HLAChain instance 
        :rtype: HLAChain
        """
        self._name:str=name
        name=mhcnames.parse_allele_name(name)
        # extract fields from the name
        self._hla_class=self.get_chain_class(name.gene)
        self._gene=name.gene
        self._allele_group=name.allele_family
        self._protein_group=name.allele_code

    def get_class(self)->int:
        """
        :returns: The HLA class 
        :rtype: int
        """
        return self._hla_class 
    
    def get_gene(self)->str:
        """
        :return: The gene name
        :rtype: str
        """
        return self._gene
    
    def get_allele_group(self)->str:
        """
        :return: The allele group 
        :rtype: str
        """
        return self._allele_group
    
    def get_protein_group(self)->str:
        """
        :return: The protein name
        :rtype: str
        """
        return self._protein_group
    
    def get_chain_class(self, gene_name:str)->int:
        """
        :param gene_name: the name of the gene 
        :type gene_name: str
        :return: 1 if the gene belongs to class one and 2 if it belong to class two 
        :rtype: int
        """
        if gene_name in ['A','B','C']: 
            return 1
        return 2
         
    def get_name(self)->str:
        """
        :return: The chain name
        :rtype: str
        """
        return self._name

    # define some magic number for the class 
    def __str__(self)->str:
        """
        :return: a string representation of the class
        :rtype: str
        """
        return f"""An HLA chain of class: {self.get_class()} from gene: {self.get_gene()},
                With an allele group of: {self.get_allele_group()} and a protein group of: 
                {self.get_protein_group()}.   
                """

    def __repr__(self)->str:
        return str(self)
        



"""A collection of convient classes to construct different IPTK classes 
"""
## load the modules and import other classes 
from __future__ import annotations
import os
import re
import time
from typing import List, Dict, Tuple, Union 
from IPTK.Classes.Experiment import Experiment
from IPTK.Classes.ExperimentSet import ExperimentSet
from IPTK.Classes.Database import SeqDB, GeneExpressionDB, CellularLocationDB,OrganismDB
from IPTK.Classes.Tissue import Tissue
from IPTK.Classes.HLASet import HLASet
from IPTK.Classes.Proband import Proband
from IPTK.Classes.Peptide import Peptide 
from IPTK.IO.InFunctions import (parse_mzTab_to_identification_table, 
parse_xml_based_format_to_identification_table, parse_text_table,parse_mzIdentML_to_identification_table) 
from tqdm import tqdm 
from concurrent.futures import ProcessPoolExecutor
from concurrent import futures
import multiprocessing as mp 
## Define the experiment class 
class RExperiment:
    def __init__(self,filepath:str, path2fasta:str, fileformat:str='idXML',tissue_name:str='total PMBC',
                proband_name:str='Default Proband', hla_set:List[str]=['DRB1*15:01','DRB1*15:01'],
                parser_param:Dict[str,Union[list,set,dict,int,float]]={})->RExperiment:
        """A Wrapper class for constracting an experimental dataset using user defined parameters\
            The class take care of initializing all classes and functions provided an easy-to-use interface\
            for working with immunopeptidomics data 
        Args:
            filepath (str): the path to load the input file, for example and idXML or an Identification table 
            path2fasta (str): the path to load Fasta database 
            fileformat (str, optional): type of input format, can be any of idXML, pepXML, mzTab or a CSV Table.\
                 Defaults to 'idXML'.
            tissue_name (str, optional): The name of the tissue to utilize, this is used for initializing the gene expression table\
                 Defaults to 'total PMBC'.
            proband_name (str, optional): the name of the proband from whome the data was obtained. Defaults to 'Default Proband'.
            hla_set (List[str], optional): A list of HLA alleles from whome the data was obtained. Defaults to ['DRB1*15:01','DRB1*15:01'].
            parser_param (Union[list,set,dict,int,float], optional): A list of parameters to be forwarded the file parser
        Returns:
            cExperimet: an IPTK.Class.Wrapper.Experiment class, an IPTK.Class.Experiment.Experiment can be extracted from the resutned instance using the get_experiment method 
        """
        ## Checking that the input is correct 
        if not os.path.exists(filepath):
            raise ValueError(f"The provided path for the identification file : {filepath} does not exist!!")
        if not os.path.exists(path2fasta):
            raise ValueError(f"The path to the proivded fasta file: {path2fasta}, does not exists!!!")
        if fileformat not in ['idXML', 'pepXML', 'csv','mzTab','mzid']:
            raise ValueError(f"Unknow input format, the provided format: {fileformat} is not supported, currently supported values are: {', '.join(['idXML', 'pepXML', 'csv','mzTab','mzid'])}")
        # define the data 
        self._proband = Proband(name=proband_name)  # the name of the proband 
        try: 
            self._hLASet = HLASet(hlas=hla_set) # just a place holder to represent the HLA allele, an instance of class HLASet
        except Exception as exp:
            raise RuntimeError(f"The following error was Encountered while creating an HLASet: \n{str(exp)}\n")
        try: 
            self._seqBD= SeqDB(path2fasta=path2fasta)
        except Exception as exp:
            raise IOError(f"While loading the fasta database the following error was Encountered : \n{str(exp)}\n")
        self._expresson_profile = GeneExpressionDB() # use the data on the human protein atlas @https://www.proteinatlas.org/about/download --> Normal tissue data 
        self._protein_locations = CellularLocationDB() # use the data on the human protein atlas @https://www.proteinatlas.org/about/download --> Subcellular location data
        try:
            self._tissue= Tissue(name=tissue_name,main_exp_value=self._expresson_profile, 
                        main_location=self._protein_locations) # create the tissue instance
        except Exception as exp:
            raise RuntimeError(f"While creating a tissue instance, the following error was Encountered: \n{str(exp)}\n")
        try:
            if fileformat == 'idXML':
                    input_table = parse_xml_based_format_to_identification_table(path2XML_file= filepath, path2fastaDB=path2fasta, is_idXML= True, **parser_param)
            elif fileformat == 'pepXML':
                     input_table = parse_xml_based_format_to_identification_table(path2XML_file= filepath, path2fastaDB=path2fasta, is_idXML= False, **parser_param)
            elif fileformat == 'mzTab':
                input_table = parse_mzTab_to_identification_table(path2mzTab= filepath, path2fastaDB=path2fasta, **parser_param)
            elif fileformat=='':
                input_table = parse_mzIdentML_to_identification_table(filepath)
            else:
                input_table = parse_text_table(filepath,path2fasta, **parser_param)
        except Exception as exp:
            raise ValueError(f"Loading the input table has caused to the following error: \n{str(exp)}\n")
        # constructing the experiments 
        try: 
            self._exp=Experiment(self._proband, self._hLASet, self._tissue, self._seqBD, input_table)
        except Exception as exp:
            raise RuntimeError(f"Generating an Experiment instance has caused to the following error: \n{str(exp)}\n")
        self._cashed_results=dict()
        return 
    ##  
    def get_experiment(self)->Experiment:
        """return the instance Experiment

        Returns:
            Experiment: an IPTK.Class.Experiment.Experiment that can be used for analyzing the data
        """
        return self._exp 
    def cash_results(self, results_name:str, results_value):
        """add a pair of results to the current instance hashmap  

        Args:
            results_name (str): the name of the result to add 
            results_value ([type]): the result value to add 
        """
        self._cashed_results[results_name]=results_value

class ReplicatedExperiment:
    def __init__(self,path:str, anchor_name:str ,path2fasta:str,
        fileformat:str='idXML', tissue_name:str='total PMBC', proband_name:str='Default Proband',
        hla_set:List[str]=['DRB1*15:01','DRB1*15:01'],
        parser_param:Dict[str,Union[list,set,dict,int,float]]={})->ReplicatedExperiment:
        """A Wrapper around an IPTK.Classes.Wrapper.Experiment instance, providing a consise, easy-to-use,  
        interface to compare different replicates. 

        Args:
            path (Union[str,os.path]): The path to the root directory where both files are defined 
            anchor_name (str): The name shared among all replicates 
            path2fasta (str): the path to load Fasta database 
            fileformat (str, optional): type of input format, can be any of idXML, pepXML, mzTab or a CSV Table (IdTable).\
                 Defaults to 'idXML'.
            tissue_name (str, optional): The name of the tissue to utilize, this is used for initializing the gene expression table\
                 Defaults to 'total PMBC'.
            proband_name (str, optional): the name of the proband from whom the data was obtained. Defaults to 'Default Proband'.
            hla_set (List[str], optional): A list of HLA alleles from whom the data was obtained. Defaults to ['DRB1*15:01','DRB1*15:01'].
            parser_param (Union[list,set,dict,int,float], optional): A list of parameters to be forwarded the file parser.

        Returns:
            PairedExperiment: [description]
        """
        ## Check that the provided path exist 
        if not os.path.exists(path):
            raise ValueError(f"The provided root directory: {path}, does not exist")
        ## get path to samples replicates 
        replicates=[os.path.join(path, name) for name in os.listdir(path) if re.match(anchor_name, name) ]
        ## check the correct number of replicates 
        if len(replicates)==0:
            raise ValueError(f"Searching using the provided path: {path} with the provided anchor name: {anchor_name} returned zero files")
        elif len(replicates)==1:
            raise ValueError(f"Searching using the provided path: {path} with the provided anchor name: {anchor_name} returned only one file: {replicates[0]}!")
        ## define a dict to hold the replicates 
        exps=dict()
        try:
            for file_name in tqdm(replicates):
                parsed_name=file_name.split('/')[-1].split('.')[0]
                exps[parsed_name]=RExperiment(file_name, path2fasta, fileformat, tissue_name, proband_name, hla_set,parser_param=parser_param).get_experiment()
        except Exception as exp:
            raise RuntimeError(f"Creating an Experimental instance from the file: {file_name} failed with the following error \n{str(exp)}\n")
        self._exps=ExperimentSet(**exps)
        self._cashed_results=dict()
        return  
    
    def get_repro_rate(self, level='protein')->float:
        """Compute the reproducibility rate among the replicates through Jaccard Index
        Args:
            level (str, optional): The level of comparison, current version support, protein and peptide. Defaults to 'protein'.

        Raises:
            ValueError: If the level is not a peptide or protein

        Returns:
            float: The reproducibility rate among the replicates. 
        """
        if level not in ['peptide','protein']:
            raise ValueError(f"The provided reproducibility level: {level} is not supported, supported levels are: peptide and protein levels.")
        if level=='protein':
            return len(self.get_concise_proteins())/len(self.get_union_protein())
        elif level=='peptide':
            return len(self.get_concise_peptides())/len(self.get_union_peptides())
        
    def get_union_protein(self)->List[str]:
        """return a list of all proteins defined among the replicates, i.e. the union accross all replicates 

        Returns:
            List[str]: a list of protein ids identified among all experiments 
        """
        return self._exps.get_unique_proteins()

    def get_concise_proteins(self)->List[str]:
        """ return the list of proteins IDss for that have been identified in all replicates 
        Returns:
            List[str]: [description]
        """
        return self._exps.get_proteins_present_in_all()

    def get_union_peptides(self)->List[str]:
        """ return a list of all peptides detected among replicates, i.e. the union accross all replicates 
        Returns:
            List[str]: a list of all peptides identified among all replicates
        """
        return self._exps.get_unique_peptides()
         
    def get_concise_peptides(self)->List[str]:
        """ return the list of peptide IDss for that have been identified in all replicates 
        Returns:
            List[str]: [description]
        """
        return self._exps.get_peptides_present_in_all()

    def get_experiments(self)->ExperimentSet:
        """
        Returns:
            ExperimentSet: the instance ExperimentSet object containing all the elements 
        """
        return self._exps 

    def cash_results(self, result_name:str,result_value:object):
        """append results to the instance hash map of results.
        Args:
            result_name (str): the name
            result_value (object): the results to be associated with the provided name 
        """
        self._cashed_results[result_name]=result_value
    
class RExperimentSet:
    def __init__(self,path:str,path2fasta:List[str],
        fileformat:List[str]=['idXML'], tissue_name:List[str]=['total PMBC'],
        proband_name:List[str]=['Default Proband'],
        hla_set:List[List[str]]=[['DRB1*15:01','DRB1*15:01']],
        num_worker:int=mp.cpu_count(),
        parser_param:Dict[str,Union[list,set,dict,int,float]]={}
        )->RExperimentSet:
        """Construction an ExperimentSet from all the input identification table

        Args:
            path (Union[str,os.path]): The path to the root directory. 
            path2fasta (str): the path to load Fasta database 
            fileformat (str, optional): type of input format, can be any of idXML, pepXML, mzTab or a CSV Table (IdTable).\
                Defaults to 'idXML'.
            tissue_name (str, optional): The name of the tissue to utilize, this is used for initializing the gene expression table\
                Defaults to 'total PMBC'.
            proband_name (str, optional): the name of the proband from whome the data was obtained. Defaults to 'Default Proband'.
            hla_set (List[str], optional): A list of HLA alleles from whome the data was obtained. Defaults to ['DRB1*15:01','DRB1*15:01'].
            parser_param (Union[list,set,dict,int,float], optional): A list of parameters to be forwarded the file parser.
        Raises:
            ValueError: in case no file could not be found in the root directory or if the length of parameter mismatched

        Returns:
            RExperimentSet: An RExperimentSet containing a collection of experiments each of which correspond to a file found in the provided root directory 
        """
        ## Check that the path exists 
        if not os.path.exists(path):
            raise ValueError(f"The provided root directory: {path} does not exists!")
        # Get the filenames relative to the path base 
        filenames=[os.path.join(path,name) for name in os.listdir(path) if fileformat in name] 
        # Check that the path exists  
        if len(filenames)==0:
            raise ValueError(f"Could not find any file in the provided directory: {path} with the provided format: {fileformat}") 
        # build the collection of experiments 
        list_jobs=[]
        print(f"Creating an ExperimentSet from: {len(filenames)} files using: {num_worker}, starting at: {time.ctime()}")
        # submitting the jobs to a process pool 
        with ProcessPoolExecutor(num_worker) as exc:
            for idx in range(len(filenames)):
                parsed_name=filenames[idx].split('/')[-1].split('.')[0]
                list_jobs.append(exc.submit(build_experiments,parsed_name,filenames,path2fasta,
                fileformat, tissue_name, proband_name, hla_set,parser_param
                ))
        success_process=0
        exps=dict()
        for fut in futures.as_completed(list_jobs):
            name,experiment=fut.result()
            exps[name]=experiment
            success_process+=1
            print(f"Progress: {success_process} files have been read and parsed into experiment, progress is: {(success_process/len(filenames))*100}%")
        ## create instance resources 
        self._exps=ExperimentSet(exps)
        self._cashed_results=dict()
        return

    def get_experiments(self)->ExperimentSet:
        """ 
        Returns:
            ExperimentSet: The instance ExperimentalSet 
        """
        return self._exps

    def cash_results(self, result_name:str,result_value:object):
        """append results to the instance hash map of results.
        Args:
            result_name (str): the name
            result_value (object): the results to be associated with the provided name 
        """
        self._cashed_results[result_name]=result_value
        
class ReplicatedExperimentSet:
    def __init__(self,path:str,path2fasta:List[str],
        fileformat:List[str]=['idXML'], tissue_name:List[str]=['total PMBC'],
        proband_name:List[str]=['Default Proband'],
        hla_set:List[List[str]]=[['DRB1*15:01','DRB1*15:01']],
        num_worker:int=mp.cpu_count(),
        parser_param:Dict[str,Union[list,set,dict,int,float]]={})->ReplicatedExperimentSet:
        """
        Args:
            path (Union[str,os.path]): The path to the root directory where files are located
            path2fasta (str): the path to load Fasta database 
            fileformat (str, optional): type of input format, can be any of idXML, pepXML, mzTab or a CSV Table (IdTable).\
                 Defaults to 'idXML'.
            tissue_name (str, optional): The name of the tissue to utilize, this is used for initializing the gene expression table\
                 Defaults to 'total PMBC'.
            proband_name (str, optional): the name of the proband from whome the data was obtained. Defaults to 'Default Proband'.
            hla_set (List[str], optional): A list of HLA alleles from whome the data was obtained. Defaults to ['DRB1*15:01','DRB1*15:01'].            
            parser_param (Union[list,set,dict,int,float], optional): A list of parameters to be forwarded the file parser.

        Raises:
            ValueError: in case no file could not be found in the root directory or if the length of parameter mismatched

        Returns:
            ReplicatedExperimentSet: a collection of replicated Experiments. 
        """
        ## Check that file name exists 
        if not os.path.exists(path):
            raise ValueError(f"The provided path: {path}, does not exists")
        ## list file names:
        filenames=[os.path.join(path,name) for name in os.listdir(path) if fileformat in name] 
        ## load the file names 
        if len(filenames)==0:
            raise ValueError(f"Could not find any file in the provided directory: {path} with the provided format: {fileformat}")
        ## get experimental names:
        replicates=self._group_paired_files(filenames)
        ## allocate a container to hold the results 
        list_jobs=[]
        ## Create an executioner to submit results too it 
        print(f"Creating an ExperimentSet from: {len(filenames)} files using: {num_worker}, starting at: {time.ctime()}")
        with ProcessPoolExecutor(num_worker) as exc:
            for replicate in replicates:
                anchor_name=self._get_common_prefix(replicate[0],replicate[1])
                list_jobs.append( exc.submit(
                    build_repeated_experiments,path,anchor_name,path2fasta, fileformat, tissue_name, proband_name, hla_set,parser_param
                ))
        self._results=dict()  
        success_process=0
        for fut in futures.as_completed(replicates):
            anchor_name,rep_exp=fut.result()
            self._results[anchor_name]=rep_exp
            success_process+=1
            print(f"Progress: {success_process} files have been read and parsed into experiment, progress is: {(success_process/len(replicates))*100}%")
        return 

    def _group_paired_files(self, file_names:List[str])->List[List[str]]:
        """ takes a list of input files and returns a list of lists where each child list contains replicate files

        Args:
            file_names (List[str]): a collection of file names 

        Returns:
            List[[str]]: a list of lists where each child list contains replicate files
        """
        results=[]
        for file_name_outer in file_names:
            for file_name_inter in file_names:
                if self._are_replicates(file_name_outer,file_name_inter):
                    results.append([file_name_outer,file_name_inter])
        # we need to make sure we have unique elements 
        for replicate in results:replicate.sort()
        temp_dev=set(['$'.join(replicate) for replicate in results])
        results=[replicate.split('$') for replicate in temp_dev]
        return results 

    def _are_replicates(self, filename_1:str, filename_2:str)->bool:
        """a helper function that check whether two files are replicate or not be comparing their names

        Args:
            filename_1 (str): The filename of the first experiment 
            filename_2 (str): The filename of the second experiment 

        Returns:
            bool: True if the two filenames are identical strings that only differ in one charachter and False otherwise. 
        """
        if len(filename_1)!=len(filename_2):return False 
        if filename_1==filename_2: return False 
        diff=[]
        index_mismatch=[]
        counter=0
        for a,b in zip(filename_1,filename_2):
            if a!=b:
                diff.append([a,b])
                index_mismatch.append(counter)
            if len(diff)>1:
                return False
            counter+=1
        if diff[0][0] not in '123456789' and diff[0][1] not in '123456789':
            return False
        if filename_1[index_mismatch[0]+1]!='_' or filename_2[index_mismatch[0]+1]!='_':
            return False
        return True 

    def _get_common_prefix(self,file_name1:str,file_name2:str)->str:
        """take an equal length file names and return the shared prefix between both files 

        Args:
            file_name1 (str): The name of the first file 
            file_name2 (str): The name of the second file 

        Returns:
            str: the prefix string between both files 
        """
        if len(file_name1)!=len(file_name2):
            raise ValueError(f"Provided String MUST be of equal length, found first file to have a length of {en(file_name1)} while the second has a length of {en(file_name2)}")
        idx=0
        file_name1=file_name1.split('/')[-1]
        file_name2=file_name2.split('/')[-1]
        for a, b in zip(file_name1,file_name2):
            if a!=b:
                break
            idx+=1
        return file_name1[:idx]+'(1|2|3|4|5|6|7|8|9)'+file_name1[idx+1:]

    def get_experiments(self)->Dict[str,ReplicatedExperiment]:
        """
        Returns:
            Dict[str,ReplicatedExperiment]: a dict containg a collection of ReplicatedExperiments  
        """
        return self._results
## Define helper analysis function use for multiprocessing 
#---------------------------------------------------------
def build_experiments(parsed_name:str, file_name:str, path2fasta:str, fileformat:str,
                    tissue_name:str, proband_name:str, hla_set:str)->Tuple[str,Experiment]:
    """[summary]

    Args:
        parsed_name (str): The name of the experiment in the results dictionary 
        file_name (str): The path to load the identification table. 
        path2fasta (str): The path to load the fasta identification files
        fileformat (str): type of input format, can be any of idXML, pepXML, mzTab or a CSV Table (IdTable).\
                 Defaults to 'idXML'.
        tissue_name (str): The name of the tissue to utilize, this is used for initializing the gene expression table\
                 Defaults to 'total PMBC'.
        proband_name (str): the name of the proband from whome the data was obtained. Defaults to 'Default Proband'.
        hla_set (str): A list of HLA alleles from whome the data was obtained. Defaults to ['DRB1*15:01','DRB1*15:01'].   

    Returns:
        Experiment: An experimental object constructed from the provided parameter 
    """
    return parsed_name, RExperiment(file_name, path2fasta, fileformat, tissue_name, proband_name, hla_set).get_experiment()
###========================================================================================================================
def build_repeated_experiments(path:str, anchor_name:str, path2fasta:str, fileformat:str,
                    tissue_name:str, proband_name:str, hla_set:str)->Tuple[str,Experiment]:
    """Build a new ReplicatedExperiment instance 

    Args:
        parsed_name (str): The name of the experiment in the results dictionary 
        file_name (str): The path to load the identification table. 
        path2fasta (str): The path to load the fasta identification files
        fileformat (str): type of input format, can be any of idXML, pepXML, mzTab or a CSV Table (IdTable).\
                 Defaults to 'idXML'.
        tissue_name (str): The name of the tissue to utilize, this is used for initializing the gene expression table\
                 Defaults to 'total PMBC'.
        proband_name (str): the name of the proband from whome the data was obtained. Defaults to 'Default Proband'.
        hla_set (str): A list of HLA alleles from whome the data was obtained. Defaults to ['DRB1*15:01','DRB1*15:01'].   

    Returns:
        ReplicatedExperimentSet: a collection of replicated Experiments. 
    """
    return anchor_name, ReplicatedExperiment(path, anchor_name, path2fasta, fileformat, tissue_name, proband_name, hla_set)




#!/usr/bin/env python 
"""An ExperimentSet which is a collection of experiments.
The class provides an API for integrating and comparing different experimental instances.   
"""
# load the modules 
from __future__ import annotations
import time
from tqdm import tqdm 
import numpy as np 
import pandas as pd
from IPTK.Classes.Experiment import Experiment 
from IPTK.Classes.Peptide import Peptide
from IPTK.Analysis.AnalysisFunction import (get_binnary_peptide_overlap, get_binnary_protein_overlap, 
    compute_jaccard_index, compute_change_in_protein_representation,compute_expression_correlation)
from typing import Dict, List
## define some types 
Experiments=Dict[str,Experiment]
Names=List[str]
Counts=Dict[str,int ]
Peptides=List[Peptide]
Proteins=List[str]
# define the class 
class ExperimentSet: 
    """an API for integrating and comparing different experimental instances
    """
    def __init__(self,**exp_id_pair)->ExperimentSet:
        """Create a new instance using an arbitrary number of experiments. 
        """
        # an experimental dict 
        self._exps=dict()
        # if the number of the provided experiments is zero, we create an an empty instance that will be filled later 
        if len(exp_id_pair)==0: return 
        # get the key and value from the exp-id-pair 
        for ident, exp in exp_id_pair.items():
            if isinstance(exp,Experiment): 
                self._exps[ident]=exp
            else:
                raise ValueError(f"""The constructor expected the value in your input pair to be an instance of class Experiment,
                                     however, your key: {ident} has a type of {type(exp)}""")
    
    def add_experiment(self, **exp_id_pair)->None:
        """adds an arbitrary number of experiments to the set 
        """
        for ident, exp in exp_id_pair.items():
            if isinstance(exp,Experiment): 
                self._exps[ident]=exp
            else:
                raise ValueError(f"""The constructor expected the value in your input pair to be an instance of class Experiment,
                                     however, your key: {ident} has a type of {type(exp)}""")

    def get_num_experiments_in_the_set(self)->int:
        """
        :return: The number of experiments currently in the set 
        :rtype: int
        """
        return len(self._exps)
    
    def get_experiments(self)->Dict[Experiment]:
        """
        :return: returns a dict with all the experiments stored in the instance as value and ids as keys. 
        :rtype: Dict[Experiment]
        """
        return self._exps
    
    def get_experiment(self, exp_name:str)->Experiment:
        """ returns the experiment pointed to by the provided experimental name

        :param exp_name: the name of the experiment 
        :type exp_name: str
        :raises KeyError: if the provided experimental name is not in the dataset. 
        :return: the experiment corresponds to the provided name 
        :rtype: Experiment
        """
        # define the experiment name 
        if exp_name not in self._exps.keys():
            raise KeyError(f"The provided experimental name: {exp_name} is not defined in the current experiment") 
        # return the experiment 
        return self._exps[exp_name]

    def get_experimental_names(self)->Names:
        """
        :return: A list with all the identifiers of the experiments in the set  
        :rtype: Names
        """
        return list(self._exps.keys())
    
    def get_unique_orgs(self)->List[str]:
        """
        :return: A list of the unique organisms in the set
        :rtype: List[str]
        """
        unique_orgs: List[str] = []
        for name in self.get_experimental_names():
             unique_orgs.extend(self._exps[name].get_orgs())
        # compute the unique organisms in the list of organisms 
        unique_orgs=list(set(unique_orgs))
        return unique_orgs

    def get_total_peptide_per_org_count(self) ->pd.DataFrame:
        """
        :return: The total count of peptides per organism accross the all experiments in the set. 
        :rtype: pd.DataFrame
        """
        # first, get the unique organisms in the set.  
        unique_orgs: List[str] = self.get_unique_orgs()
        # create a counter and initialize it to zero to hold the results 
        org_counter=dict()
        for org in unique_orgs:
            org_counter[org]=0
        # update the counts 
        for name in tqdm(self.get_experimental_names()):
            for _, row in self._exps[name].get_peptides_per_organism().iterrows():
                 org_counter[row['Organisms']]+=row['Counts']
       	# make the data compatible with data frames 
        for org in org_counter.keys():
            org_counter[org]=[org_counter[org]]
		# create a dataframe
        res: pd.DataFrame = pd.DataFrame(org_counter).T
		# add the index as an extra-columns 
        res['Organisms'] = res.index.tolist()
        res.columns=['Counts','Organisms']
        # reformat the dataframe 
        res.reset_index(drop=True,inplace=True)
        res=res.reindex(columns=['Organisms','Counts'])
        # sort the results 
        res=res.sort_values(by='Counts',ascending=False)
        return res

    def compare_org_count_among_exps(self, org:str, abs_count: bool =False) ->pd.DataFrame: 
        """
        :param org: The name of the organism to query the database for it. 
        :type org: str
        :param abs_count: The absolute count, defaults to False
        :type abs_count: bool, optional
        :return: The count of the peptides that belong to a specific organism in the database.
        :rtype: pd.DataFrame
        """
        # allocate and array to hold the results 
        res: np.ndarray = np.zeros((len(self),len(self)))
        # loop over all the experiments in the set 
        # initialize the counters 
        experimental_name: List[str] = self.get_experimental_names()
        for row_idx in tqdm(range(len(experimental_name))):
            # get the counts per column 
            org_row: pd.DataFrame = self._exps[experimental_name[row_idx]].get_peptides_per_organism()
            org_row_count: int = org_row.loc[org_row.iloc[:,0]==org]['Counts'] 
            if org_row_count.empty:
                org_row_count=0
            else:
                org_row_count=org_row_count.tolist()[0]
            # get the counts per row 
            for col_idx in range(len(experimental_name)):
                # get the row-column
                org_col: pd.DataFrame = self._exps[experimental_name[col_idx]].get_peptides_per_organism()
                org_col_count: int = org_col.loc[org_col.iloc[:,0]==org]['Counts'] 
                if org_col_count.empty: 
                    org_col_count=0
                else: 
                    org_col_count=org_col_count.tolist()[0]
                # add the count to the
                res[row_idx,col_idx]= org_row_count-org_col_count
        # return a data frame of the results 
        res=pd.DataFrame(res)
        # add the name of columns and index 
        res.columns=experimental_name
        res.index=experimental_name
        # return the results
        return res 
    
    def drop_peptides_belong_to_org(self, org_name: str) -> None:
        """drops all the peptides that belong to the provided organisms from all experiments in the set.


        :param org_name: The name of the organism to drop.
        :type org_name: str
        """
        for name in self.get_experimental_names():
            self[name].drop_peptide_belong_to_org(org_name)
        return 

    def __getitem__(self, name:str)->Experiment:
        """ A magic function for accessing the experiments stored in the set

        :param name: The experiment name or id 
        :type name: str
        :raises KeyError: if the provided name is not defined in the current instance. 
        :return: the experiment with the corresponding name
        :rtype: Experiment
        """
        if name in self.get_experimental_names():
            return self._exps[name]
        else:
            raise KeyError(f"Name: {name} is not defined in the current instance")
    
    def __len__(self)->int:
        """  A magic function for computing the length of the set, which is the number of experiments 
        stored inside the instance 
        
        :return: The number of experiments stored inside the instance 
        :rtype: int 
        """
        return len(self._exps)
    
    def __str__(self)->str:
        """ A string representation of the class

        :return: A string representation of the class
        :rtype: str
        """
        return f'an experimental set with {len(self)} Experiments in it.'
        
    
    def get_tissue_counts(self)->Counts:
        """
        :return: The number of experiments obtained from each tissue in the current instance
        :rtype: Counts
        """
        tissues_counts=dict()
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            tissue_=self._exps[exp].get_tissue_name()
            if tissue_ in tissues_counts.keys():
                tissues_counts[tissue_]+=1
            else: 
                tissues_counts[tissue_]=1
        # return the results 
        return tissues_counts
    
    def get_allele_count(self)->Counts:
        """
        :return: The number of experiments obtained from each allele in the instance.
        :rtype: Counts
        """
        allele_counts=dict()
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            alleles=self._exps[exp].get_hla_allele()
            for allele in alleles:
                if allele in allele_counts.keys():
                    allele_counts[allele]+=1
                else: 
                    allele_counts[allele]=1
        # return the results 
        return allele_counts
    
    def get_proband_count(self)->Counts:
        """
        :return: The number of experiments obtained from each proband in the ExperimentSet.
        :rtype: Counts
        """
        proband_count=dict()
        # loop over all the elements in the set 
        for exp in self._exps.keys(): 
            proband_=self._exps[exp].get_proband_name()
            if proband_ in proband_count.keys():
                proband_count[proband_]+=1
            else: 
                proband_count[proband_]=1
        # return the gene_counts 
        return proband_count
    
    def group_by_tissue(self)->Dict[str,ExperimentSet]:
        """
        :return:  A map between each tissue and an ExperimentSet object representing all experiments belonging to that tissue. 
        :rtype: Dict[str,ExperimentSet]
        """
        # define the set of tissues 
        tissues2exps: Dict[str,ExperimentSet]=dict()
        tissue_counter: Dict[str, int]=dict()
        # Initialize the counters 
        tissues: List[str] =list(set([self._exps[exp].get_tissue_name() for exp in self._exps.keys()]))
        for tissue in tissues:
            tissue_counter[tissue]=0
        # loop over all the elements in the set 
        for exp in tqdm(self._exps.keys()): 
            tissue_=self._exps[exp].get_tissue_name()
            if tissue_ in tissues2exps.keys():
                temp_pair: Dict[str,Experiment]={
                    tissue_+str(tissue_counter[tissue_]):self._exps[exp]
                }
                tissues2exps[tissue_].add_experiment(**temp_pair)
                tissue_counter[tissue_]+=1
            else:
                temp_pair: Dict[str,Experiment]={
                    tissue_+str(tissue_counter[tissue_]):self._exps[exp]
                } 
                tissues2exps[tissue_]=ExperimentSet(**temp_pair)
                tissue_counter[tissue_]+=1
        # return results 
        return tissues2exps
        
    def group_by_proband(self)->Dict[str,ExperimentSet]:
        """
        :return: a map between each proband and an Experimentset object representing all experiments objects \
        belonging to each unique proband in the set. 
        :rtype: Dict[str,ExperimentSet]
        """
        # define the set of tissues 
        proband2exps: Dict[str, ExperimentSet]=dict()
        proband_counter: Dict[str,int]=dict() 
        # define the probands 
        probands: List[str] = list(set([self._exps[exp].get_proband_name() for exp in self._exps.keys()]))
        for proband in probands: 
            proband_counter[proband]=0
        # loop over all the elements in the set 
        for exp in tqdm(self._exps.keys()): 
            proband_=self._exps[exp].get_proband_name()
            if proband_ in proband2exps.keys():
                temp_pair: Dict[str,Experiment]={
                    proband_+str(proband_counter[proband_]):self._exps[exp]
                }
                proband2exps[proband_].add_experiment(**temp_pair)
                proband_counter[proband_]+=1
            else:
                temp_pair: Dict[str,Experiment]={
                    proband_+str(proband_counter[proband_]):self._exps[exp]
                }
                proband2exps[proband_]=ExperimentSet(**temp_pair)
                proband_counter[proband_]+=1
        # return results 
        return proband2exps

    
    def get_unique_peptides(self)->Peptides:
        """
        :return: The set of unique peptides in the experimentSet 
        :rtype: Peptides
        """
        res:List[str]=[]
        for exp_name in self.get_experimental_names(): 
            res.extend(self[exp_name].get_peptides())
        return list(set(res))
    
    def get_unique_proteins(self)->Proteins:
        """
        :return: The set of unique proteins in the experimentset
        :rtype: Proteins
        """
        res:List[str]=[]
        for exp in self.get_experimental_names():
            res.extend(self[exp].get_proteins())
        return list(set(res))
    
    def is_peptide_present_in_all(self, peptide:str)->bool:
        """
        :param peptide: The peptide sequence to search its occurrences in every experiment contained in the set
        :type peptide: str
        :return: True if peptide is present in all experiments inside the instance, False otherwise 
        :rtype: bool
        """
        # check whether the peptide is defined in the set of unique peptides or not  
        if peptide not in self.get_unique_peptides():
            return False
        # check if the peptide is a member of ALL EXPERIMENTs in the set 
        for exp_name in self.get_experimental_names(): 
            if not self[exp_name].is_member(peptide):
                return False 
        # if the peptide is a member of every experiment in the set, we return True
        return True
    
    def is_protein_present_in_all(self, protein:str)->bool:
        """
        :param protein: the protein id to search its occurrences in every experimental in the set 
        :type protein: str
        :return: True if peptide is present in all experiments inside the instance, False otherwise 
        :rtype: bool
        """
        if protein not in self.get_unique_proteins():
            return False
        #check that the protein is a member of every experiment 
        for exp_name in self.get_experimental_names():
            if not self[exp_name].is_a_parent_protein(protein):
                return False
        # return true as it is defined in all the experiment in the dataset. 
        return True

    def get_peptides_present_in_all(self)->Peptides:
        """
        :return: The peptides that are observed in every experiment in the set.
        :rtype: Peptides
        """
        all_peptides=self.get_unique_peptides()
        results=[]
        for pep in all_peptides: 
            if self.is_peptide_present_in_all(pep):
                results.append(pep)
        return results 
    
    def get_proteins_present_in_all(self)->Proteins:
        """
        :return: The proteins that are inferred in all experiments of the set. 
        :rtype: Proteins
        """
        all_proteins=self.get_unique_proteins()
        results: List[str]=[]
        for protein in all_proteins: 
            if self.is_protein_present_in_all(protein):
                results.append(protein)
        return results 
        
    def compute_peptide_overlap_matrix(self, method:str='count')->np.ndarray:
        """

        :return: A 2D matrix containing the number of peptide overlapping between each pair of experiments inside the current instance collection of experiment.   
        :rtype: np.ndarray
        :param method: The method of computing the overlap methods, can be count or jaccard, incase of count, exact matches between samples are used, i.e. number of overlaps,\
             meanwhile, incase of jaccard, jaccard Index between each pair of sample is used.  
        :type protein: str
        """
        # validate input method: 
        #----------------------
        if method!='count' and method!='jaccard':
            raise ValueError(f"Method: {method} is not supported!, only count and jaccard are currently supported.")
        # allocate the results array 
        results_array=np.zeros(shape=(len(self), len(self))) 
        experiment_names=self.get_experimental_names()
        for raw_idx in tqdm(range(len(experiment_names))):
            for col_idx in range(len(experiment_names)):
                if method=='count':
                    results_array[raw_idx,col_idx]=len(get_binnary_peptide_overlap(self[experiment_names[raw_idx]],self[experiment_names[col_idx]])) # compute the peptide overlap    
                elif method=='jaccard':
                    results_array[raw_idx,col_idx]=compute_jaccard_index(self[experiment_names[raw_idx]],self[experiment_names[col_idx]],level='peptide')
        # construct a dataframe from the results 
        results_df=pd.DataFrame(results_array)      
        # add the columns and index to the results df 
        results_df.columns=experiment_names
        results_df.index=experiment_names
        # return the results 
        return results_df 

    def compute_protein_overlap_matrix(self,method:str='count')->pd.DataFrame:
        """
        :return: returns a 2D matrix containing the number of proteins overlapping between each pair of experiments inside the current instance collection of experiment.  
        :rtype: np.ndarray
        :param method: The method of computing the overlap methods, can be count or jaccard, incase of count, exact matches between samples are used, i.e. number of overlaps,\
             meanwhile, incase of jaccard, jaccard Index between each pair of sample is used.  
        :type protein: str
        """
        # validate input method: 
        #----------------------
        if method!='count' and method!='jaccard':
            raise ValueError(f"Method: {method} is not supported!, only count and jaccard are currently supported.")
        # allocate the results array 
        results_array=np.zeros(shape=(len(self), len(self))) 
        experiment_names=self.get_experimental_names()
        for raw_idx in tqdm(range(len(experiment_names))):
            for col_idx in range(len(experiment_names)):
                if method=='count':
                    results_array[raw_idx,col_idx]=len(get_binnary_protein_overlap(self[experiment_names[raw_idx]],self[experiment_names[col_idx]])) # compute the peptide overlap    
                elif method=='jaccard':
                    results_array[raw_idx,col_idx]=compute_jaccard_index(self[experiment_names[raw_idx]],self[experiment_names[col_idx]],level='protein')
        # construct a dataframe from the results 
        results_df=pd.DataFrame(results_array)      
        # add the columns and index to the results df 
        results_df.columns=experiment_names
        results_df.index=experiment_names
        # return the results 
        return results_df 
    
    def compute_peptide_representation_count(self)->Counts: 
        """
        :return: The number of times a peptide was observed accross experiments in the set 
        :rtype: Counts
        """
        # define the results object 
        results: Dict[str, int]=dict()
        unique_peptides: List[str]=self.get_unique_peptides()
        # fill the dictionary and initialize the counts to zeros 
        for peptide in unique_peptides:
            results[peptide]=0
        # loop over all the experiment to count the peptides 
        for peptide in tqdm(unique_peptides):
            for exp_name in self.get_experimental_names():
                if self[exp_name].is_member(peptide):
                    results[peptide]+=1
        # return the results after filling it 
        return results 
    
    def compute_protein_representation_count(self)->Counts: 
        """
        :return: The number of times a protein was observed accross the experiment in the set
        :rtype: Counts
        """
        results=dict()
        unique_proteins=self.get_unique_proteins()
        # fill the dictionary and initialize the counts to zeros 
        for prot in unique_proteins:
            results[prot]=0
        # loop over all the experiments to count the peptides 
        for prot in tqdm(unique_proteins):
            for exp_name in self.get_experimental_names():
                if self[exp_name].is_a_parent_protein(prot):
                    results[prot]+=1
        # return the results after filling it 
        return results  
    
    def compute_protein_coverage_over_the_set(self)->Dict[str, np.ndarray]:
        """
        :return: The mapped representation for each protein accross the entire set
        :rtype: Dict[str, np.ndarray]
        """
        results:Dict[str, np.ndarray]=dict()
        unique_proteins=self.get_unique_proteins()
        # get a consent representation 
        for prot in tqdm(unique_proteins):
            for exp_name in self.get_experimental_names():
                if self[exp_name].is_a_parent_protein(prot):
                    if prot in results.keys(): 
                        results[prot]+=self[exp_name].get_mapped_protein(prot)
                    else:
                        results[prot]=self[exp_name].get_mapped_protein(prot)
        # return the results after filling the array 
        return results


    def compute_correlation_in_experssion(self)->pd.DataFrame:
        """computes the correlation in parent protein gene-expression across all the experiments
        in the set. See the function **compute_binary_correlation** in the analysis module 
        for information about the computational logic.

        :return: returns a 2D matrix containing the coorelation in gene expression between each pair of experiments inside the current instance collection of experiments.  
        :rtype: pd.DataFrame
        """
        # get the experimental index
        exps_ids: List[str] = self.get_experimental_names()  
        # allocate an array to hold the results
        res_array: np.ndarray = np.zeros((len(exps_ids),len(exps_ids)))
        # fill the array with the expression correlation 
        for row_idx in tqdm(range(len(exps_ids))): 
            for col_idx in range(len(exps_ids)):
                res_array[row_idx,col_idx] = compute_expression_correlation(
                    self.get_experiment(exps_ids[row_idx]),
                    self.get_experiment(exps_ids[col_idx])
                )
        # create a dataframe 
        res: pd.DataFrame =pd.DataFrame(res_array)
        # add the col and row names 
        res.columns=exps_ids
        res.index=exps_ids
        # return the results 
        return res 

    def compute_change_in_protein_representation(self)->np.ndarray:
        """Computes the change in protein representation among the proteins that are presented/ detected in all of the \
        instance's experiments.
        
        :returns: a 3D tensor, T, with shape of (num-experiments, num-experiments, num-proteins), \
        where T[i,j,k] is a the difference between experiment i & j with respect to the kth protein \
        :rtype: np.ndarray
        """
        # get the number of experiments and proteins 
        present_in_all: List[str] = self.get_proteins_present_in_all()
        num_exps: int = self.get_num_experiments_in_the_set()
        # allocate an array to hold the results 
        results_array: np.ndarray = np.zeros(shape=(num_exps, num_exps,len(present_in_all)))
        ## fill the array 
        # create some counters 
        col_counter: int = 0
        row_counter: int= 0
        for prod_idx in tqdm(range(len(present_in_all))): 
            for exp_col in self.get_experiments().keys():
                for exp_row in self.get_experiments().keys():
                    results_array[row_counter,col_counter,prod_idx]=compute_change_in_protein_representation(
                            self.get_experiment(exp_row).get_mapped_protein(present_in_all[prod_idx]), 
                            self.get_experiment(exp_col).get_mapped_protein(present_in_all[prod_idx])
                        )
                    # increase the row counters 
                    row_counter+=1
                row_counter=0
                # increase the columns counter
                col_counter+=1 
            col_counter=0
        # return the results
        return results_array

    def compute_average_distance_between_exps(self)->pd.DataFrame: 
        """computes the average distance between experiments by taking the average over the z-axis
        of the 3D tensor computed by the function compute_change_in_protein_representation.
        
        :return:  A 2D tensor with shape of (num-experiments, num-experiments)
        :rtype: pd.DataFrame
        """
        diff_protein_overlap: np.ndarray = self.compute_change_in_protein_representation()
        # average overlap is: 
        average_scores: np.ndarray = np.sum(diff_protein_overlap,axis=-1)
        # construct a dataframe of the results 
        results_df: pd.DataFrame = pd.DataFrame(average_scores)
        # add the colnames 
        results_df.columns=list(self.get_experiments().keys())
        results_df.index=list(self.get_experiments().keys())
        # return the results 
        return results_df
    
    def compare_peptide_counts(self)->pd.DataFrame:
        """
        :return: A table that contain the total number of peptides and per-organism peptide counts \
        among all experiments in the set
        :rtype: pd.DataFrame
        """
        # allocate a list to hold the results 
        total_count: List[str]= []
        experiment_name: List[str]= []
        # fill the lists 
        for name in self.get_experimental_names():
            experiment_name.append(name)
            total_count.append(len(self[name].get_peptides()))
        # create a dataframe to store the results 
        res: pd.DataFrame = pd.DataFrame({
            'Organisms':['Total']*len(total_count), 
            'Counts':total_count,
            'Names':experiment_name
        })
        # get the count per organism in each experiment 
        for exp_name in self.get_experimental_names():
            # get the temp results 
            temp_res: pd.DataFrame = self[exp_name].get_peptides_per_organism()
            # add the experiment name as an extra column 
            temp_res['Names'] = [exp_name] * temp_res.shape[0]
            # concatenate the results
            res= pd.concat([res,temp_res], axis=0)
        # return the results 
        return res
    
    def compute_peptide_length_table(self)->pd.DataFrame:
        """
        :return: A table that contain the length of each peptide in the experiment
        :rtype: pd.DataFrame
        """
        # allocate a data frame to hold the results 
        res: pd.DataFrame = pd.DataFrame(columns=['Peptide_length','Names'])
        # loop over all the experiment
        for exp_name in self.get_experimental_names():
            temp_res: pd.DataFrame = pd.DataFrame({
                'Peptide_length': self[exp_name].get_peptides_length(),
                'Names': [exp_name]*len(self[exp_name].get_peptides_length())
            }) 
            res= pd.concat([res, temp_res],axis=0)
        # return the results 
        return res

    def get_num_peptide_per_experiment(self)->pd.DataFrame:
        """return a table containing the number of peptides in every experiment in the current set.

        :return: A table with two columns, the first is the experiment name and the second is the peptide count per experiment.
        :rtype: pd.DataFrame
        """
        results:pd.DataFrame= pd.DataFrame({ exp_name:[len(exp.get_peptides())] for exp_name, exp in self._exps.items()},
                                index=['experiment','num_peptide']).T
        results.sort_values(by='num_peptide', inplace=True, ascending=False)
        return results

    def get_num_proteins_per_experiment(self)->pd.DataFrame:
        """return a table containing the number of proteins in every experiment in the current set.

        :return: A table with two columns, the first is the experiment name and the second is the protein count per experiment.
        :rtype: pd.DataFrame
        """
        results:pd.DataFrame= pd.DataFrame({ exp_name:[len(exp.get_proteins())] for exp_name, exp in self._exps.items()},
                                index=['experiment','num_proteins']).T
        results.sort_values(by='num_peptide', inplace=True, ascending=False)
        return results
    
   

















#!/usr/bin/env 
""" TandomSpectra is a simple class to link MS1 and MS2 spectra 
"""
## LOAD THE MODULES
#------------------
from __future__ import annotations
import pyopenms as poms 
## DEFINe THE CLASS
#------------------
class TandomSpectra:
    def __init__(self,parent:poms.pyopenms_4.MSExperiment,
        children:List[poms.pyopenms_4.MSExperiment])->TandomSpectra:
        """class constructor, construct a new instance of class TandomSpectra

        :param parent: an MS1 spectrum that acts as the precursor for the class  
        :type parent: poms.pyopenms_4.MSExperiment
        :param children: MS2 spectra obtained from the precursor MS1 spectrum 
        :type children: a list of MS2 spectrums 
        :return: a new instance 
        :rtype: TandomSpectra
        """
        if parent.getMSLevel()!=1:
            raise ValueError(f'Parent Spectrum must have an MS-level of 1, however, current parent has: {parent.getMSLevel()}')
        else: 
            self._parent=parent
        for child in children:
            if child.getMSLevel() != 2:
                raise ValueError(f'All the children spectra must be of level 2, however, spectra: {child.getNativeID()} has a level of: {child.getMSLevel()}')
        self._children=children
        return 
    
    def __len__(self)->int:
        """a magic function to return the number of children in the class 

        :return: [description]
        :rtype: int
        """
        return len(self._children)
    
    def __repr__(self)->str:
        return str(self)

    def __str__(self)->str:
        return f'a TandomSpectra with {len(self)} children'
        


#!/usr/bin/env python
"""An interface to query and retrive data from an database designed according to the builtin database, see IPTK.Utils.CreateDB for more details 
"""
## LOAD THE MODULES 
#------------------
from __future__ import annotations
import pandas as pd 
import os
from typing import List
from datetime import datetime
import mhcnames
## DEFINE THe CLASS
#------------------
class BuildInDB:
    def __init__(self,path2file:str, sep:str='\t')->BuildInDB:
        """Loads a pandas dataframe that comply with the expected structure

        :param path2file: The path to load the CSV files. 
        :type path2file: string. 
        :param sep: The seperator for the table, defaults to tab. 
        :type sep: string. 
        :return: a buildIn database instance that can be used to handle the sequences in the database
        :rtype: BuildInDB
        """
        try:
            self._database=pd.read_csv(path2file,sep=sep)
        except Exception as exp:
            raise IOError(f'while reading the database located in: {path2file} the following error was encountered: {str(exp)}')
        try:
            self._validate_database(self._database)
        except Exception as exp:
            raise ValueError(f'The provided data is not a valid database, because: {str(exp)}')
    
    def _validate_database(self, database:pd.DataFrame)->None:
        """ validate that a provided database has the correct number, name and order of columns.

        :param database: a proposed database to validate 
        :type database: pd.DataFrame
        :raises ValueError: incase validating the database failed 
        """
        if database.shape[1]!=11:
            raise ValueError(f'Expected the provided table to have 11 eleven columns, however the provided table has : {self._database.shape[1]}')
        def_col_name=['experiment_id','peptide','protein','start_pos','end_pos','organism','allele','source','bioSample','antibody','search_engine']
        if database.columns!=def_col_name:
            raise ValueError(f"Expected the provided table to have the exact order and name of columns as follow, {', '.join(def_col_name)}, however the \
                the current table has {', '.join(database.columns)}.")
        return
    
    def __str__(self)-> str:
        return f'A build in database with {self._database.shape[0]} entries.'

    def __repr__(self) -> str: 
        return str(self)
    
    def __len__(self)->int:
        return self._database.shape[0]

    def append_to_data_base(self,content_to_add: pd.DataFrame)->None:
        """[summary]

        :param content_to_add: [description]
        :type content_to_add: pd.DataFrame
        """
        try:
            self._validate_database(content_to_add) 
        except ValueError as exp:
            raise ValueError(f'Validating your provided table because: {str(exp)}')
        self._database=pd.concat([self._database,content_to_add], axis=0)
        return 

    def write_to_disk(self, path:str, add_timestampe:bool = True)->None:
        """write the generated database to the disk, the default saperator used is tab

        :param path: the path, INCLUDING the name of the genrated database 
        :type path: str
        :param add_timestampe: wether or not to a timestamp to the end of the file_name, defaults to True
        :type add_timestampe: bool, optional
        """
        if add_timestampe:
            time_stamp=str(datetime.now()).replace(' ','_')  
            path=path.split('.')[0]
            path+=time_stamp
        try:
            self._database.to_csv(path+'.tsv',sep='\t',index=False)
        except Exception as exp:
            raise IOError(f'While writing the database the following error was encountered: {exp}')
        return

    def get_peptides_from_experiment(self,experiment_id:str)->List[str]:
        """returns a list of peptides that were identified from a specific experiment

        :param experiment_id: the experiment id 
        :type experiment_id: str
        :raises ValueError: if the experiment id is not defined in the database 
        :return: a list of peptides that were identified from a specific experiment
        :rtype: List[str]
        """
        if experiment_id not in set(list(self._database.iloc[:,0])):
            raise ValueError(f'The provided id is not in the database')
        return self._database.loc[self._database.iloc[:,0]==experiment_id].iloc[:,1].tolist() 

    def get_proteins_from_experiment(self,experiment_id:str)->List[str]:
        """returns a list of proteins that were identified from a specific experiment.

        :param experiment_id: the experiment id 
        :type experiment_id: str
        :raises ValueError: if the experiment id is not defined in the database 
        :return: a list of proteins that were identified from a specific experiment
        :rtype: List[str]
        """
        if experiment_id not in set(list(self._database.iloc[:,0])):
            raise ValueError(f'The provided id is not in the database')
        return self._database.loc[self._database.iloc[:,0]==experiment_id].iloc[:,2].tolist() 

    def get_table_from_experiment(self,experiment_id:str)->pd.DataFrame:
        """returns a subtable contain all info associated with the provided experimental-id 

        :param experiment_id: the experiment id 
        :type experiment_id: [type]
        :raises ValueError: [description]
        :return: a table contain all the experiments associated with the provided experimental-id 
        :rtype: pd.DataFrame
        """
        if experiment_id not in set(list(self._database.iloc[:,0])):
            raise ValueError(f'The provided id is not in the database')
        return self._database.loc[self._database.iloc[:,0]==experiment_id]

    def get_peptides_bound_by_allele(self,allele_name:str)->pd.DataFrame:
        """ returns a subtable contain all info associated with the provided allele_name

        :param allele_name: the allele name written in the standard notation 
        :type allele_name: str
        :raises ValueError: incase the provided allele is not in the database 
        :return: a subtable contain all the experiments associated with the provided allele_name
        :rtype: pd.DataFrame
        """
        s_allele_name=mhcnames.parse_allele_name(allele_name)
        s_allele_name=s_allele_name.gene+'*'+s_allele_name.allele_family+':'+s_allele_name.allele_code
        current_alleles=self.get_all_unique_alleles_in_db()
        if s_allele_name not in current_alleles:
            raise ValueError(f"The provided allele is not currently defined in the database, currently the following is defined: {' ,'.join(current_alleles)}")
        return  self._database.loc[self._database.allele.contains(s_allele_name),]
    
    def get_all_unique_alleles_in_db(self)->List[str]:
        """ returns a list of all unique alleles in the database
        :return: a list of all unique alleles in the database
        :rtype: List[str]
        """
        allele_pairs=list(set([alleles for alleles in self._database.allele]))
        alleles = []
        for allele_pair in allele_pairs:
            alleles.extend(allele_pair.extend(';')) 
        return list(set(alleles))
    
    def get_all_unique_proteins_in_db(self)->List[str]:
        """ return a list of all unique peptides in the database 

        :return: all unique peptides in the database 
        :rtype: List[str]
        """
        return list(set(self._database.iloc[:,1]))


    def get_all_unique_peptides_in_db(self)->List[str]:
        """ return a list of all unique proteins in the database 

        :return: all unique proteins in the database 
        :rtype: List[str]
        """
        return list(set(self._database.iloc[:,2]))
 
    def get_all_unique_organims_in_db(self)->List[str]:
        """ return a list of all unique proteins in the database 

        :return: all unique proteins in the database 
        :rtype: List[str]
        """
        return list(set(self._database.iloc[:,5]))

    def get_organism_info_from_db(self, org_name:str)->pd.DataFrame:
        """returns a sub-table contain all info associated with the provided org_name

        :param org_name: the name of the organism to query the database 
        :type org_name: string 
        :return: a sub-table contain all info associated with the provided org_name
        :rtype: pd.DataFrame
        """
        current_organims=self.get_all_unique_organims_in_db()
        if org_name not in current_organims:
            raise ValueError(f'The provided organims is not in the database: {org_name}')
        return self._database.loc[self._database.organism==org_name,]

    def get_peptide_info_from_db(self,peptide_seq)->pd.DataFrame:
        """returns a sub-table contain all info associated with the provided peptide

        :param peptide_seq: the peptide sequence to query the database with 
        :type peptide_seq: string 
        :return: a sub-table contain all info associated with the provided org_name
        :rtype: pd.DataFrame
        """ 
        return self._database.loc[self._database.peptide==peptide_seq,]

    def get_protein_info_from_db(self,protein_id)->pd.DataFrame:
        """returns a sub-table contain all info associated with the provided protein id

        :param protein_id: protein id
        :type peptide_seq: string 
        :return: a sub-table contain all info associated with the provided protein id 
        :rtype: pd.DataFrame
        """ 
        return self._database.loc[self._database.protein==protein_id,] 
        
    def get_table(self)->pd.DataFrame:
        return self._database
    
    


"""The class provides wrapper for performing GO analysis for the identified protein hits, the engine is wrapper around the 
GOATools library.   
"""
## import the modules
from __future__ import annotations 
import os 
import time 
from typing import List, Set, Dict, Union
import pandas as pd
from IPTK.Classes.DataFeeders.gene_ncbi_9606_prot_encoding import GENEID2NT as gene2iden_human
from IPTK.Classes.DataFeeders.gene_ncbi_10090_prot_encoding import GENEID2NT as gene2iden_mouse 
from IPTK.Utils.Mapping import map_from_uniprot_to_Entrez_Gene
from IPTK.Classes.Experiment import Experiment 
import tqdm as tqdm 
from goatools.base import download_go_basic_obo,download_ncbi_associations
from goatools.obo_parser import GODag
from goatools.anno.genetogo_reader import Gene2GoReader
from goatools.goea.go_enrichment_ns import GOEnrichmentStudyNS
from goatools.go_enrichment import GOEnrichmentRecord

## define the engine class 
class GOEngine:
    def __init__(self, work_dir: str='.', clean_work_dir: bool = False,
                organism:str='human',
                study_parameters:Dict[str,Union[int,float,str,List,Dict]]={
                    'propagate_counts':False,
                    'alpha':0.05, 
                    'methods':['fdr_bh']
                })->GOEngine:
        """A GOEngine that can be used for performing analysis using GOATOOLS

        Args:
            work_dir (str, optional): The path to a temp directory were intermediate-results and raw data will be downloaded/written to. Defaults to the current working directory.
            clean_work_dir (bool, optional): Whether or not to remove data written to the work directory at class termination, default to True.
            organism (str, optional): The organism . Defaults to 'human'.
            study_parameters (Dict[str,Union[int,float,str,List,Dict]], optional): A dict of parameters to control the base function, defaults to {'propagate_counts':False,'alpha':0.05, 'methods':['fdr_bh']}
        Returns:
            GOEngine: return a GO engine that can be used for performing GO enrichment analysis GOEnrichmentStudyNS
        """
        print("Creating a GO Engine ...")
        if not os.path.exists(work_dir):
            raise ValueError(f"The provided work path: {work_dir} does not exist!!!")
        self.work_dir=work_dir
        if organism != 'human' and organism!='mouse':
            raise ValueError(f"The provided organism: {organism} is not support, current engine mainly work with human and moues only")
        print(f"\t --> Downloading data ...")
        obo_fname=download_go_basic_obo(os.path.join(work_dir,'go-basic.obo'))
        gene2go_fname=download_ncbi_associations(os.path.join(work_dir,'gene2go'))
        ## parse the GO term 
        print(f"\t --> parsing the data and intializing the base GOEA object...")
        obo_dag=GODag(obo_fname)
        if organism=='human':
            self._goea_obj=GOEnrichmentStudyNS(
                gene2iden_human.keys(), 
                Gene2GoReader(gene2go_fname,taxids=[9606]).get_ns2assc(),
                obo_dag,
                **study_parameters)
        else:
            self._goea_obj=GOEnrichmentStudyNS(
                gene2iden_human.keys(), 
                Gene2GoReader(gene2go_fname,taxids=[10090]).get_ns2assc(),
                obo_dag,
                **study_parameters)
        self._clean_work_dir=clean_work_dir
        self._gene_ids=None
        return 
    
    def load_data(self,exp:Experiment, num_proteins:int=-1)->None:
        """Load the data to the Engine, so GOEA can be conducted 

        Args:
            exp (Experiment): An Experimental object to extract uniprot ids 
            num_proteins (int, optional): The number of proteins to be included in the analysis. Defaults -1 to which mean use all proteins,\
                 otherwise it uses the number of proteins provided by the user. note that the function is sorted by number of peptides per protein,\
                      that is the first 10 protein means, getting the top 10 protein with most peptides. 
        Raises:
            ValueError: if the function called while data being already associated with the engine from a previous call
        """
        if self._gene_ids is not None:
            raise ValueError(f"There some data still in the engine, the first 10 genes are: {','.join(self._gene_ids[:10])}\
                clean your engine from previous data using the function, clean_engine and try again.")
        print(f"Getting the number of peptide per protein ..., started at: {time.ctime()}")
        num_protein_per_peptides=exp.get_peptides_per_protein()
        if num_proteins == -1: 
            list_proteins=num_protein_per_peptides.iloc[:,0].to_list()
        else:
            list_proteins=num_protein_per_peptides.iloc[:,0].to_list()[:num_proteins]
        print(f"Map uniprot to Entrez gene ids ..., starting at: {time.ctime()}")
        self._gene_ids= [int(gene_id) for gene_id in map_from_uniprot_to_Entrez_Gene(list_proteins).iloc[:,1].to_list()]
        print(f"{len(self._gene_ids)} Genes have been correctly loaded")
        return 

    def run_analysis(self, quite:bool=False, only_signifcant:bool=True,
            significance_level:float=0.05, get_list_term:bool=False)->Union[pd.DataFrame, List[GOEnrichmentRecord]]:
        if quite:
            goea_results=self._goea_obj.run_study(self._gene_ids, prt=None)
        else:
            goea_results=self._goea_obj.run_study(self._gene_ids)
        if only_signifcant:
            goea_results=[res for res in goea_results if res.p_fdr_bh<significance_level]
        if get_list_term:
            return goea_results
        else: 
            self._goea_obj.wr_tsv(os.path.join(self.work_dir,'temp_file.tsv'),goea_results)
            results_df=pd.read_csv(os.path.join(self.work_dir,'temp_file.tsv'),sep='\t')
            os.system(f"rm -f {os.path.join(self.work_dir,'temp_file.tsv')}")
        return results_df

    def clean_engine(self)->None:
        """Remove Current list of gene ids associated with the engine 
        """
        self._gene_ids=None
        return 

    def __del__(self)->None:
        """class destructor, clean work directory if  clean_work_dir is set to True 
        """
        if self.clean_work_dir: os.system(f"rm -f {self.work_dir}/*")
        return 





""""
A configuration wrapper most of the library parameters 
"""
from __future__ import annotations
from typing import Dict,Set,List,Union
import matplotlib as mpl 
class ParamConfig:
    def __init__(self,output_directory:str=None,
        matplot_lib_param:Dict[str,Union[int,float,str]]=None, 
        plotly_lib_param:Dict[str,Union[int,float,str]]=None)-> ParamConfig:
        """Create a new parameter configuration, which library global parameters 

        Args:
            matplot_lib_param (Dict[str,Union[int,float,str]], optional): a dict of parameters that controls global matplotlib parameters. Defaults to None.
        Returns:
            ParamConfig: [description]
        """
        self.params["matplot_lib_param"]=matplot_lib_param,
        return

    def set(self)->None:
        if self.params["matplot_lib_param"]  is not None:
            mpl.rcParams.update(self.params["matplot_lib_param"])


    def update(self,param_dict:Dict[str,Union[Dict[str,Union[int,float,str]],int,float,str]])->None:
        """Takes a dict to update the parameters of the config dict 

        Args:
            param_dict (Dict[str,Union[Dict[str,Union[int,float,str]],int,float,str]]): the new configuration parameters 
        """
        self.params.update(param_dict) 



#!/usr/bin/env python 
"""Write the results generated by the library into a wide variety of formats. 
"""
import numpy as np 
import h5py
from IPTK.Classes.Experiment import Experiment
from IPTK.Classes.Peptide import Peptide
from typing import List
import pandas as pd
# define some types 
Peptides= List[Peptide]
Names= List[str]
# define the function of the module 
def write_mapped_tensor_to_h5py(tensor: np.ndarray, path2write: str, dataSet_name: str ='MAPPED_TENSOR')->None:
    """ Write a mapped tensor to an hdf5 file 

    :param tensor: The provided tensor to  write it to the hdf5 file. 
    :type tensor: np.ndarray
    :param path2write: The path of the output file 
    :type path2write: str
    :param dataSet_name: The name of the dataset inside the mapped tensor, defaults to 'MAPPED_TENSOR'
    :type dataSet_name: str, optional
    :raises IOError: In case opening the file for writing failed 
    """
    try:
        file_handler=h5py.File(path2write,'w')
    except Exception as exp:
        raise IOError(f'While opening the file: {path2write} for writing, the following error was encountered: {exp}')
    # create the dataset 
    file_handler.create_dataset(dataSet_name,data=tensor)
    # flush the data 
    file_handler.close()
    return 

def write_annotated_sequences(peptides: List[str], labels: List[int], path2write: str,
                            sep: str = ',', shuffle: bool =True) ->None:
    """take a list of peptides along with its sequences and write the results to a CSV file.
    
    :param peptides: A list of peptide sequences 
    :type peptides: List[str]
    :param labels: A list of numerical labels associated with the peptides 
    :type labels: List[int]
    :param path2write: the path to write the generated file 
    :type path2write: str
    :param sep: The separator in the resulting table, defaults to ','
    :type sep: str, optional
    :param shuffle: Whether or not to shuffle the table, defaults to True
    :type shuffle: bool, optional
    :raises ValueError: incase the length of the tables and labels is not matching
    :raises IOError: In case writing the output table failed
    """
    if len(peptides) != len(labels):
        raise ValueError(f'The provided lists do not match, peptides have: {len(peptides)} elements while labels have: {len(peptides)} elements')
    # define the dataframe results 
    table: pd.DataFrame = pd.DataFrame({
        'peptides':peptides,
        'labels': labels
    }) 
    # shuffle the database 
    if shuffle:
        table=table.sample(frac=1.0)
    # write the results 
    try: 
        table.to_csv(path2write,sep=sep, index=False, header=True)
    except Exception as exp: 
        raise IOError(f'While writing the results table, the following error was encountered: {exp}')
    

def write_named_peptides_to_fasta(names: Names, peptides: Peptides, output_file:str):
    """Takes a list of names and peptide sequences and writes them as an output file to the disk as fasta files

    :param names: A list of sequences names 
    :type names: Names
    :param peptides: A list of peptide sequences 
    :type peptides: Peptides
    :param output_file: The name of the output file to write the results to, it will OVERWRITE existing files
    :type output_file: str
    :raises ValueError: Incase the length of the tables and labels is not matching
    :raises IOError: In case writing the output file failed
    """
    if len(names) != len (peptides): 
        raise ValueError(f"The length of the names container: {len(names)} does not equal the length of the sequences container: {len(peptides)}")
    # write the files 
    try:
        with open(output_file, 'w') as writer:
            for idx in range(len(names)): 
                writer.write('>'+names[idx]+'\n')
                writer.write(peptides[idx]+'\n')
    except Exception as exp: 
        raise IOError(f'While writing the output file the following error was encountered: {exp}') 

def write_auto_named_peptide_to_fasta(peptides:Peptides, output_file:str )->None:
    """ Takes a list of peptides, generate automatic names for the peptides and write the results to the disk as FASTA files 

    :param peptides: a list of peptide sequences 
    :type peptides: Peptides
    :param output_file: the name of the output file to write the results to, it will OVERWRITE existing files
    :type output_file: str
    """
    # generate automatic names for the pepetide 
    names=['PEP_AUTO_NAME_'+str(idx) for idx in range(len(peptides))]
    # call the writer functions 
    write_named_peptides_to_fasta(names, peptides,output_file)
    
def write_pep_file(peptides:Peptides, output_file:str):
    """Takes a file and write the peptides to .pep file which is a text file that contain one peptide per line
    
    :param peptides: a list of peptide sequences 
    :type peptides: Peptides
    :param output_file: the name of the output file to write the results to, it will OVERWRITE existing files
    :type output_file: str
    :raises IOError: In case writing the output file failed
    """
    try: 
        with open(output_file,'w') as writer: 
            for pep in peptides: 
                writer.write(pep+'\n')
    except Exception as exp: 
        raise IOError(f'While writing the file, the following error was encountered: {exp}')
    


#/usr/bin/env python 
"""A collection of assembled analysis functions that automate different immunopeptidomics analysis task 
"""
## load the modules 
from IPTK.Classes.Experiment import Experiment
import numpy as np 
import pandas as pd 
import itertools
from tqdm import tqdm 
import time 
from typing import Dict, List, Set

def compute_protein_coverage(experiment1:Experiment,experiment2:Experiment,progress_bar:bool=True)->Dict[str,Dict[str,np.ndarray]]:
    """ Compute the difference in protein coverage among the two input experiments

    Args:
        experiment1 (Experiment): The first experiment containing the protein and peptides derived from the first HLA-set 
        experiment2 (Experiment): The second experiment containing the protein and peptides derived from the second HLA-set
        progress_bar (bool): A boolean flag for controlling the progress bar, if true, a progress bar is shown, defaults to True. 

    Returns:
        Dict[str,Dict[str,np.ndarray]]: Returns a nested dict containing protein identifiers as a keys and a dict as a value, the dict contain two arrays as values,\
            the first contain protein coverage in the first HLA-Set and the second contain the coverage in the second HLA-set.\
                  The results dictionary only contain coverage for proteins observed in the two sets. 
    """
    protein_experiment_one:List[str]=experiment1.get_proteins()
    protein_experiment_two:List[str]=experiment2.get_proteins()
    present_in_both:List[str]=protein_experiment_one.intersection(protein_experiment_two)
    results:Dict[str,Dict[str,np.ndarray]]=dict()
    if progress_bar:
        for protein in present_in_both:
            temp_dict={
                '_'.join(experiment1.get_hla_set().get_names()):experiment1.get_mapped_protein(protein),
                '_'.join(experiment2.get_hla_set().get_names()):experiment2.get_mapped_protein(protein)
            }
            results.update({protein:temp_dict})
    else:
        for protein in present_in_both:
            temp_dict={
                '_'.join(experiment1.get_hla_set().get_names()):experiment1.get_mapped_protein(protein),
                '_'.join(experiment2.get_hla_set().get_names()):experiment2.get_mapped_protein(protein)
            }
            results.update({protein:temp_dict})
    return results



"""Impelement Numba JIT accelerated functions for scaling up IPTK code
"""
## load the modules 
from numba.np.ufunc import parallel
import numpy as np
from numba import jit
from typing import List, Dict, Set, Union

@jit(nopython=True,nogil=True,parallel=True,cache=True)
def _get_mapped_proteins(start_idx:np.ndarray, end_idx:np.ndarray, protin_len:int)->np.ndarray:
    """Compute mapped protein array 

    Args:
        start_idx:(np.ndarray): the start boundires of child peptide 
        end_idx (np.ndarray): the end boundires of child peptide 
        protin_len (int): the length of the parent protein

    Returns:
        np.ndarray: an np.ndarray of shape(1,protein_len) containing the mapping of peptides to protein
    """
    protein_backbone=np.zeros(shape=(1,protin_len),dtype=np.int32)
    for i,j in zip(start_idx,end_idx): protein_backbone[0,i:j]+=1
    return protein_backbone

@jit(nopython=True,nogil=True,parallel=False,cache=True)
def _get_non_presented_peptides(exc_reg_s_idx: int, exc_reg_e_idx: int, length: int, seq:str)->str:
    """ get a non-presented peptides from the protein backbone 

    Args:
        exc_reg_s_idx (int): the start position of the represented peptide 
        exc_reg_e_idx (int): the end position of the represented peptide
        length (int): the length of the non-represented peptide 
        seq (str): the protin sequence 

    Returns:
        str: a non-presented peptide which is a subsequence of the protein sequence 
    """
    try_flag=True
    while try_flag: 
        anc_point=np.random.randint(low=0,high=len(seq)-length)
        if anc_point < exc_reg_s_idx or anc_point > exc_reg_e_idx: 
            neg_seq=seq[anc_point:anc_point+length] 
            try_flag=False
    return neg_seq


#!/usr/bin/env Python 
"""An interface to query and retrive data from AFND
"""
## LOAD THE MODULES 
#------------------
from __future__ import  annotations
import urllib3
import mhcnames
import pandas as pd 
from bs4 import BeautifulSoup
##DEFINE THE CLASS
#-----------------
class AFND:
    def __init__(self)->AFND:
        """class constructor  

        :return: an AFND instance that can be used to query data from the website
        :rtype: AFND
        """
        self._base='http://www.allelefrequencies.net/hla6002a.asp?all_name='
        self._http=urllib3.PoolManager()
        return

    def get_frequency_of(self,allele_name:str)->pd.DataFrame:
        """The worker function of the class, Queries AFND for the frequency of the provided allele in different populations

        :param allele_name: the allele name in standard notation, for example, A*01:01 or DRB1*15:01
        :type allele_name: string
        :return: a tuples of two columns, the first contain the country name and the second contains the frequency
        :rtype: pd.DataFrame
        """
        try:
            allele_name=mhcnames.parse_allele_name(allele_name)
        except mhcnames.AlleleParseError as exp: 
            raise ValueError(f'While parsing the provided name: {allele_name} the following error was encountered: {str(exp)}')
        q_name = allele_name.gene+'*'+allele_name.allele_family+':'+allele_name.allele_code
        try: 
            print(f'sending the query to the server ....')
            response=self._http.request('GET',self._base+q_name) 
            print(f'response received')
        except Exception as exp: 
            raise RuntimeError(f'While querying the server the following error was encountered: {str(exp)}')
        # check that the response is valid 
        if response.status==404:
            raise RuntimeError(f'The provided allele list is not defined in the database, server returned 404 ')        
        # get the frequency table from the HTML page 
        html_soup=BeautifulSoup(response.data.decode('utf-8',errors='ignore'),'html.parser')
        html_table=html_soup.body.find_all("table")[3].find_all("tr")[1:]
        ## extract the data from all elements of the table 
        rows=[]
        for row in html_table:
            columns=[]
            for elem in row:
                try:
                    columns.append(elem.get_text())
                except AttributeError:
                    continue
            rows.append(columns)
        ## create the table 
        temp_table=pd.DataFrame(data=rows)
        freq_table=pd.concat([temp_table.iloc[:,1],temp_table.iloc[:,3]],axis=1)
        freq_table.columns=['Country','Frequency']
        return freq_table
    
    
   
    


#!/usr/bin/env python 
"""A representation of a protein that has been inferred from an IP experiment.  
"""
# load the modules:
from __future__ import annotations 
import numpy as np 
import pandas as pd
import IPTK.Utils.AcceleratedFunction as axf 

# define some types 
class Protein: 
	"""representation of a protein that has been infered from an IP experiment. 
	"""
	def __init__(self, prot_id : str, seq: str, org: str=None) -> Protein:
		"""construct a Protein instance. 

		:param prot_id: the id of the protein 
		:type prot_id: str
		:param seq:  the sequence of the protein 
		:type seq: str
		:param org: the organism from which the protein originated, defaults to None
		:type org: str, optional
		:return: a Protein instance
		:rtype: Protein
		"""
		self._id=prot_id
		self._seq=seq
		self._org=org
		
	def get_id(self) -> str:
		"""
		:return: return the protein identifier.
		:rtype: str
		"""
		return self._id 
	
	def get_seq(self) -> str:
		"""
		:return: the protein sequence.
		:rtype: str
		"""
		return self._seq
	
	def get_peptides_map(self, start_idxs: np.ndarray, end_idxs:np.ndarray)->np.ndarray:
		"""compute a coverage over the protein sequence
		
		:param start_idxs: a numpy array of integers representing the start positions
		:type start_idxs: np.ndarray 
		:param end_idxs: a numpy array of integers representing the end positions
		:type start_idxs: np.ndarray 
		:raises ValueError: if start_indxs and end_idxs are not of equal length. 
		:return: A numpy array with shape of 1 by the length of the protein where every element in the array donates the number of times, It has been observed in the experiment.
		:rtype: np.ndarray
		"""
		if len(start_idxs) != len(end_idxs):
			raise ValueError(f"""Annotation list MUST have an equal length where your lists have length : {len(start_idxs)}, {(end_idxs)} respectively.""")
		## check arrays have the correct type 
		if not isinstance(start_idxs,np.ndarray):start_idxs=np.array(start_idxs,dtype=np.int32)
		if not isinstance(end_idxs,np.ndarray):end_idxs=np.array(end_idxs,dtype=np.int32)
		## call the accelerated function  
		return axf._get_mapped_proteins(start_idxs,end_idxs,len(self)) 
	
	def get_non_presented_peptide(self, exc_reg_s_idx: int, exc_reg_e_idx: int, length: int) -> str: 
		"""Sample a peptide from the protein sequences where the sampled peptides are not part of the \
		experimentally identified regions. 
		
		:param exc_reg_s_idx: the start index of experimentally identified peptide, i.e. real peptide. 
		:type exc_reg_s_idx: int
		:param exc_reg_e_idx: the end index of the experimentally identified peptide, i.e. real peptide. 
		:type exc_reg_e_idx: int
		:param length: length the non-presented peptide, i.e. not-experimentally identified. 
		:type length: int
		:raises ValueError: if the length of the peptide is bigger than the protein length 
		:raises ValueError: if the length of the peptide is smaller than or equal to zero 
		:return: a substring of the instance sequence
		:rtype: str
		"""
		if length >= len(self):
			raise ValueError(f'''The generated peptide must be shorter than the parent protein length,
		 	Provided length: {length}  While, the parent length: {len(self)}''')
		if length <=0:
			raise ValueError(f'The generated peptide length MUST be  a positive integer, your input value is: {length}')
		return axf._get_non_presented_peptides(exc_reg_s_idx,exc_reg_e_idx,length,self._seq)
	
	def get_org(self)->str:
		"""
		:return: the organism in which this instance protein belong.
		:rtype: str
		"""
		return self._org

	def set_org(self,org: str)->None:
		"""A post-instantitation mechanism to set the organism for which the protein belong.
	
		:param org: the name of the organism
		:type org: str
		"""
		self._org=org

	def __getitem__(self,aa_index: int ) -> str:
		"""
		:param aa_index:  the amino acid index, for example, 3rd amino acid in the protein 5th amino acid in the protein
		:type aa_index: int
		:return: the amino acid with the corresponding index as the user provided index
		:rtype: str
		"""
		return self._seq[aa_index] 
	
	def __len__(self)->int:
		"""
		:return: the length of the protein sequence  
		:rtype: int
		"""
		return len(self._seq)
	
	def __str__(self)->str:
		"""
		:return: the protein sequence 
		:rtype: str
		"""
		return self._seq

	def __repr__(self)->str:
		return f'A protein instance with a length of {len(self)}.'


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="iptkl", 
    version="0.6.6",
    author="Hesham ElAbd",
    author_email="h.elabd@ikmb.uni-kiel.de",
    description="IPTK is a library specialized in the analysis of HLA-peptidomes identified through an immunopeptidomics pipeline.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        'Documentation': 'https://iptk.readthedocs.io/en/latest/',
        'Funding': 'https://www.genes-environment-inflammation.de', 
        'Tutorials': 'https://github.com/ikmb/iptoolkit/tree/master/Tutorials',
        'Tracker': 'https://github.com/ikmb/iptoolkit/issues',
    },
    url='https://github.com/ikmb/iptoolkit',
    packages=setuptools.find_packages(),
    install_requires=['pandas','biopython','seaborn' ,'matplotlib', 'plotly' ,'mhcnames', 'pyteomics', 'h5py', 'logomaker', 'colour', 'lxml',
             'nglview', 'sklearn', 'scipy','statannot', 'numba', 'pyopenms','bokeh','holoviews','tqdm','numba','goatools'], 
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Intended Audience :: Science/Research",
        "Typing :: Typed"
    ],
    keywords='HLA immunopeptidomics antigen-processing antigen-presentation computational-immunology interactive-data-analysis MHC',
    python_requires='>=3.7',
)



#!/usr/bin/env Python 
"""The module contains functions to to call meme software via a system call.   
"""
# load the module 
from Bio import motifs 
import subprocess as sp 
import os
from IPTK.Utils.UtilityFunction import append_to_calling_string
## define the interface functions 
def is_meme_callable()->bool:
    """
    :return: True if meme is callable, False otherwise. 
    :rtype: bool
    """
    try: 
        res=sp.run(['meme'],stdout=sp.DEVNULL, stderr=sp.DEVNULL)
        return True
    except FileNotFoundError: 
        return False

def get_meme_help()->None:
    """Print the command line help interface for the meme tool

    :raises FileNotFoundError: if meme is not callable 
    """
    if is_meme_callable():
        sp.run(['meme','-h'])
    else: 
        raise FileNotFoundError(f"meme is either not installed or not part of the PATH ==> {os.environ['PATH']}")

def call_meme(input_fasta_file:str, output_dir:str, verbose: bool=True, objfunc: str='classic', test: str = 'mhg',
              use_llr: bool = False, shuf: int = 2, hsfrac: float = 0.5, cefrac: float = 0.25, 
              searchsize: int= -1, maxsize: int = -1 , norand: bool = False, csites: int = -1, seed: int = -1,
              mod: str='oops', nmotifs: int= -1, evt: float = -1.0, time: int = -1, nsite: int = -1, 
              minsites: int = -1, maxsite: int = -1, nsites: int = -1, w: int = -1, minw: int = -1, 
              maxw: int = -1, nomatrim: bool = False, wg: int = -1, ws: int = -1, noendgaps: bool = False,
              maxiter: int = -1, prior: str = 'dirichlet', b: int = -1, p: int = -1 )->None:
    """warper for making a system call to meme software for sequence motif finding \
        for the reset of the function parameters use the function \
            **get_meme_help** defined in the module IO, submodule MEMEInterface. \ 
            
    :param input_fasta_file: The path to input FASTA files. 
    :type input_fasta_file: str
    :param output_dir: the output dir to write the results, **IT WILL OVERWRITE EXISTING DIRECTORY** 
    :type output_dir: str
    :param verbose:  whether or not to print the output of calling meme to the screen, default is True. 
    :type verbose: bool 
    """
    # define the calling string backbone
    calling_string=""
    # add the input dataset:
    calling_string += str(input_fasta_file)+" "
    ## adding the parameters to the calling string based upon their values 
    calling_string=append_to_calling_string('objfunc', 'classic', objfunc,calling_string,False)
    calling_string=append_to_calling_string('test', 'mhg', test, calling_string,False)
    calling_string=append_to_calling_string('use_llr', False, use_llr, calling_string,True)
    calling_string=append_to_calling_string('shuf', 2, shuf, calling_string,False)
    calling_string=append_to_calling_string('hsfrac', 0.5, hsfrac, calling_string,False)
    calling_string=append_to_calling_string('cefrac', 0.25, cefrac, calling_string,False)
    calling_string=append_to_calling_string('searchsize', -1, searchsize, calling_string,False)
    calling_string=append_to_calling_string('maxsize', -1, maxsize, calling_string,False)
    calling_string=append_to_calling_string('norand', False, norand, calling_string,True)
    calling_string=append_to_calling_string('csites', -1, csites, calling_string,True)
    calling_string=append_to_calling_string('seed', -1, seed, calling_string,False)
    calling_string=append_to_calling_string('mod', 'oops', mod, calling_string, False)
    calling_string=append_to_calling_string('nmotifs', -1, nmotifs, calling_string,False)
    calling_string=append_to_calling_string('evt', -1, evt, calling_string,False)
    calling_string=append_to_calling_string('time', -1, time, calling_string,False)
    calling_string=append_to_calling_string('nsite', -1, nsite, calling_string,False)
    calling_string=append_to_calling_string('minsites', -1, minsites, calling_string,False)
    calling_string=append_to_calling_string('maxsite', -1, maxsite, calling_string,False)
    calling_string=append_to_calling_string('nsites', -1, nsites, calling_string,False)
    calling_string=append_to_calling_string('minw', -1, minw, calling_string,False)
    calling_string=append_to_calling_string('maxw', -1, maxw, calling_string,False)
    calling_string=append_to_calling_string('nomatrim', False, nomatrim, calling_string,True)
    calling_string=append_to_calling_string('wg', -1, wg, calling_string,False)
    calling_string=append_to_calling_string('ws', -1, ws, calling_string,False)
    calling_string=append_to_calling_string('maxiter', -1, maxiter, calling_string,False)
    calling_string=append_to_calling_string('prior', 'dirichlet', prior, calling_string,False)
    calling_string=append_to_calling_string('noendgaps', False, noendgaps, calling_string,True)
    calling_string=append_to_calling_string('b', -1, b, calling_string,False)
    calling_string=append_to_calling_string('p', -1, p, calling_string,False)
    # add the output path
    calling_string+=" -oc "+str(output_dir)
    ## call the program 
    # generate the calling command 
    calling_command=['meme']
    calling_command.extend(calling_string.split(' ')) 
    # remove empty spaces from the calling string 
    calling_command=[elem for elem in calling_command if elem != '']
    # call the porgram 
    try:
        if verbose:
            sp.run(calling_command, check=True)
        else:
            sp.run(calling_command, check=True)
    except sp.CalledProcessError as exp: 
        raise RuntimeError(f'While calling meme to compute the motif, the following was encountered: {exp}')
    




    

    

    
    
    
    
    
    
    
    
    
    
    
    
    

    


    


#!/usr/bin/env python 
""" The class provide some convenient function for computing summary stats and extracting information from\
     a collection of mzML files  
""" 
## IMPORT THE MODULES
#--------------------
from __future__ import annotations
import os
import pyopenms as poms 
from IPTK.Classes.MzMLExperiment import MzMLExperiment
import pandas as pd
from typing import List, Callable
from tqdm import tqdm 
## CLASS DEFINTION
#-----------------
class MzMLExperimentSet:
    def __init__(self, path2files:str)->MzMLExperimentSet:
        """ create a new instance of mzMLExperimentSet from all mzML files present in a provided directory 

        :param path2files: The path to a directory containing all the files that need to be loaded
        :type path2files: str
        :raises ValueError: incase the provided path does not exists 
        :raises RuntimeError: incase the creating an MzMLExperiment instance from one of the files failed
        :return: an MzMLExperimentSet instance 
        :rtype: MzMLExperimentSet
        """
        if not os.path.exists(path2files):
            raise ValueError(f'The provided path: {path2files} sequences does not exists!!')
        mzML_files=[p_file for p_file in os.listdir(path2files) if '.mzML' in p_file]
        self._files={}
        for mzML_file in tqdm(mzML_files): 
            try:
                self._files[mzML_file.split('.')[0]]=MzMLExperiment(os.path.join(path2files,mzML_file))
            except Exception as exp:
                raise RuntimeError(f'While loading: {os.path.join(path2files,mzML_file)} the following error was encountered: {str(exp)}')
        return
    
    def get_total_number_of_spectra_per_file(self)->pd.DataFrame:
        """return the number of spectra in each MS file stored in the set of MzML experiment  

        :return: a table with two columns, the first, is the experiment name and the second is the column name 
        :rtype: pd.DataFrame
        """
        # get the count of MS spectra in each file
        counts_dict=dict()
        for p_file in tqdm(self._files.keys()):
            counts_dict[p_file]=len(self._files[p_file])
        # construct a pandas dataframe from the dict
        table=pd.DataFrame().from_dict(counts_dict,orient='index')
        table['e']=table.index
        table.columns=['Count','Experiment Name']
        table.reset_index(drop=True,inplace=True)
        table.reindex(columns=['Experiment Name','Count'],inplace=True)
        return table
    
    def get_number_of_MS1_per_file(self)->pd.DataFrame:
        """return the number of MS2 spectra in each MS file stored in the set of MzML experiment 

        :return: a table containing the count from each file  
        :rtype: pd.DataFrame
        """
        counts_dict=dict()
        for p_file in tqdm(self._files.keys()):
            counts_dict[p_file]=self._files[p_file].get_num_MS2_spectra()
        # construct a pandas dataframe from the dict
        table=pd.DataFrame().from_dict(counts_dict,orient='index')
        table['e']=table.index
        table.columns=['Count','Experiment Name']
        table.reset_index(drop=True,inplace=True)
        table.reindex(columns=['Experiment Name','Count'],inplace=True)
        return table

    def get_number_of_MS2_per_file(self)->pd.DataFrame:
        """return the number of MS2 spectra in each MS file stored in the set of MzML experiment 

        :return: a table containing the count from each file  
        :rtype: pd.DataFrame
        """
        counts_dict=dict()
        for p_file in tqdm(self._files.keys()):
            counts_dict[p_file]=self._files[p_file].get_num_MS2_spectra()
        # construct a pandas dataframe from the dict
        table=pd.DataFrame().from_dict(counts_dict,orient='index')
        table['e']=table.index
        table.columns=['Count','Experiment Name']
        table.reset_index(drop=True,inplace=True)
        table.reindex(columns=['Experiment Name','Count'],inplace=True)
        return table

